@@ -1,20 +1,22 @@
-import seisbench
-import seisbench.util
-
-from abc import abstractmethod, ABC
-from pathlib import Path
-import pandas as pd
-import h5py
-import numpy as np
-from tqdm import tqdm
-import matplotlib.pyplot as plt
-import os
-import inspect
-import scipy.signal
 import copy
+import inspect
+import os
+import warnings
+from abc import ABC, abstractmethod
 from collections import defaultdict
 from collections.abc import Iterable
-import warnings
+from pathlib import Path
+from urllib.parse import urljoin
+
+import h5py
+import matplotlib.pyplot as plt
+import numpy as np
+import pandas as pd
+import scipy.signal
+from tqdm import tqdm
+
+import seisbench
+import seisbench.util
 
 
 class WaveformDataset:
@@ -75,6 +77,12 @@ class WaveformDataset:
                                   but ignore missing ones. This will raise an error if traces with different
                                   numbers of components are requested together.
     :type missing_components: str
+    :param metadata_cache: If true, metadata is cached in a lookup table.
+                           This significantly speeds up access to metadata and thereby access to samples.
+                           On the downside, this requires storing two copies of the metadata in memory.
+                           The second copy usually consumes more memory due to the less space-efficient format.
+                           Runtime differences are particularly big for large datasets.
+    :type bool:
     :param kwargs:
     """
 
@@ -88,6 +96,7 @@ class WaveformDataset:
         cache=None,
         chunks=None,
         missing_components="pad",
+        metadata_cache=False,
         **kwargs,
     ):
         if name is None:
@@ -100,11 +109,7 @@ class WaveformDataset:
         self._chunks = chunks
         if chunks is not None:
             self._chunks = sorted(chunks)
-
-            available_chunks = self.available_chunks(path)
-            for chunk in self._chunks:
-                if chunk not in available_chunks:
-                    raise ValueError(f"Dataset does not contain the chunk '{chunk}'.")
+            self._validate_chunks(path, self._chunks)
 
         self._missing_components = None
 
@@ -117,6 +122,8 @@ class WaveformDataset:
         self._component_order = None  # Target component order
         # Dict [source_component_order -> list for reordering source to target components]
         self._component_mapping = None
+        self._metadata_lookup = None
+        self._chunks_with_paths_cache = None
         self.sampling_rate = sampling_rate
 
         self._verify_dataset()
@@ -138,6 +145,9 @@ class WaveformDataset:
             tmp_metadata["trace_chunk"] = chunk
             metadatas.append(tmp_metadata)
         self._metadata = pd.concat(metadatas)
+        self._metadata.reset_index(
+            inplace=True
+        )  # Required because grouping of chunked datasets breaks otherwise
 
         self._data_format = self._read_data_format()
 
@@ -148,9 +158,18 @@ class WaveformDataset:
         self.dimension_order = dimension_order
         self.component_order = component_order
         self.missing_components = missing_components
+        self.metadata_cache = metadata_cache
 
         self._waveform_cache = defaultdict(dict)
 
+        self.grouping = None
+
+    def _validate_chunks(self, path, chunks):
+        available_chunks = self.available_chunks(path)
+        for chunk in chunks:
+            if chunk not in available_chunks:
+                raise ValueError(f"Dataset does not contain the chunk '{chunk}'.")
+
     def __str__(self):
         return f"{self._name} - {len(self)} traces"
 
@@ -214,6 +233,15 @@ class WaveformDataset:
 
         self._cache = cache
 
+    @property
+    def metadata_cache(self):
+        return self._metadata_cache
+
+    @metadata_cache.setter
+    def metadata_cache(self, val):
+        self._metadata_cache = val
+        self._rebuild_metadata_cache()
+
     @property
     def path(self):
         """
@@ -293,6 +321,43 @@ class WaveformDataset:
 
         self._component_order = value
 
+    @property
+    def grouping(self):
+        """
+        The grouping parameters for the dataset. These parameters are used to determine the
+        :py:attr:`~groups` and for the associated methods.
+        `grouping` can be either a single string or a list of strings.
+        Each string must be a column in the metadata.
+        By default, the grouping is None.
+        """
+        return self._grouping
+
+    @grouping.setter
+    def grouping(self, value):
+        self._grouping = value
+
+        if value is None:
+            self._groups = None
+            self._groups_to_trace_idx = None
+        else:
+            self._metadata.reset_index(
+                inplace=True, drop=True
+            )  # Make sure the indices are correct
+            self._groups_to_trace_idx = self.metadata.groupby(value).groups
+            self._groups = list(self._groups_to_trace_idx.keys())
+            self._groups_to_group_idx = {
+                group: i for i, group in enumerate(self._groups)
+            }
+
+    @property
+    def groups(self):
+        """
+        The list of groups as defined by the :py:attr:`~grouping` or `None` if :py:attr:`~grouping` is `None`.
+        """
+        return copy.copy(
+            self._groups
+        )  # Return a copy to make the internal groups immutable
+
     @property
     def chunks(self):
         """
@@ -362,6 +427,14 @@ class WaveformDataset:
 
         return sorted(chunks)
 
+    def _rebuild_metadata_cache(self):
+        if self.metadata_cache:
+            self._metadata_lookup = list(
+                self._metadata.apply(lambda x: x.to_dict(), axis=1)
+            )
+        else:
+            self._metadata_lookup = None
+
     def _unify_sampling_rate(self, eps=1e-4):
         """
         Unify sampling rate to common base. Operation is performed inplace. The sampling rate can be
@@ -451,7 +524,6 @@ class WaveformDataset:
         :param target:
         :return:
         """
-
         if (
             isinstance(source, float)
             and np.isnan(source)
@@ -519,10 +591,21 @@ class WaveformDataset:
 
         :return: List of chunks, list of metadata paths, list of waveform paths
         """
-        metadata_paths = [self.path / f"metadata{chunk}.csv" for chunk in self.chunks]
-        waveform_paths = [self.path / f"waveforms{chunk}.hdf5" for chunk in self.chunks]
+        if self._chunks_with_paths_cache is None:
+            metadata_paths = [
+                self.path / f"metadata{chunk}.csv" for chunk in self.chunks
+            ]
+            waveform_paths = [
+                self.path / f"waveforms{chunk}.hdf5" for chunk in self.chunks
+            ]
 
-        return self.chunks, metadata_paths, waveform_paths
+            self._chunks_with_paths_cache = (
+                self.chunks,
+                metadata_paths,
+                waveform_paths,
+            )
+
+        return self._chunks_with_paths_cache
 
     def _verify_dataset(self):
         """
@@ -606,6 +689,22 @@ class WaveformDataset:
                     "Keeping original components."
                 )
 
+    def get_group_idx_from_params(self, params):
+        """
+        Returns the index of the group identified by the params.
+
+        :param params: The parameters identifying the group. For a single grouping parameter, this argument will be a
+                       single value. Otherwise this argument needs to be a tuple of keys.
+        :return: Index of the group
+        :rtype: int
+        """
+        self._verify_grouping_defined()
+
+        if params in self._groups_to_group_idx:
+            return self._groups_to_group_idx[params]
+        else:
+            raise KeyError("The dataset does not contain the requested group.")
+
     def get_idx_from_trace_name(self, trace_name, chunk=None, dataset=None):
         """
         Returns the index of a trace with given trace_name, chunk and dataset.
@@ -731,6 +830,8 @@ class WaveformDataset:
             self._metadata = self._metadata[mask]
             self._evict_cache()
             self._build_trace_name_to_idx_dict()
+            self._rebuild_metadata_cache()
+            self.grouping = self.grouping  # Recalculate grouping
         else:
             other = self.copy()
             other.filter(mask, inplace=True)
@@ -887,7 +988,10 @@ class WaveformDataset:
         :param sampling_rate: Target sampling rate, overwrites sampling rate for dataset.
         :return: Tuple with the waveforms and the metadata of the sample.
         """
-        metadata = self.metadata.iloc[idx].to_dict()
+        if self._metadata_lookup is None:
+            metadata = self.metadata.iloc[idx].to_dict()
+        else:
+            metadata = copy.deepcopy(self._metadata_lookup[idx])
 
         if sampling_rate is None:
             sampling_rate = self.sampling_rate
@@ -945,7 +1049,11 @@ class WaveformDataset:
                 idx = [idx]
                 squeeze = True
 
-            load_metadata = self._metadata.iloc[idx]
+            if self._metadata_lookup is None:
+                load_metadata = self._metadata.iloc[idx]
+            else:
+                load_metadata = [self._metadata_lookup[i] for i in idx]
+                load_metadata = self._pack_metadata(load_metadata)
         else:
             if mask is not None:
                 load_metadata = self._metadata[mask]
@@ -996,6 +1104,80 @@ class WaveformDataset:
 
         return waveforms
 
+    def get_group_size(self, idx):
+        """
+        Returns the number of samples in a group
+
+        :param idx: Group index
+        :type idx: int
+        :return: Size of the group
+        :rtype: int
+        """
+        self._verify_grouping_defined()
+        group = self._groups[idx]
+        idx = self._groups_to_trace_idx[group]
+        return len(idx)
+
+    def get_group_samples(self, idx, **kwargs):
+        """
+        Returns the waveforms and metadata for each member of a group.
+        For details see :py:func:`get_sample`.
+
+        :param idx: Group index
+        :type idx: int
+        :param kwargs: Kwargs passed to :py:func:`get_sample`
+        :return: List of waveforms, list of metadata dicts
+        """
+        return self._get_group_internal(idx, return_metadata=True, **kwargs)
+
+    def get_group_waveforms(self, idx, **kwargs):
+        """
+        Returns the waveforms for each member of a group.
+        For details see :py:func:`get_sample`.
+
+        :param idx: Group index
+        :type idx: int
+        :param kwargs: Kwargs passed to :py:func:`get_sample`
+        :return: List of waveforms
+        """
+        return self._get_group_internal(idx, return_metadata=False, **kwargs)
+
+    def _get_group_internal(self, idx, return_metadata, **kwargs):
+        self._verify_grouping_defined()
+
+        group = self._groups[idx]
+        idx = self._groups_to_trace_idx[group]
+
+        waveforms = []
+        metadata = []
+
+        for trace_idx in idx:
+            trace_wv, trace_meta = self.get_sample(trace_idx, **kwargs)
+            waveforms.append(trace_wv)
+            metadata.append(trace_meta)
+
+        if return_metadata:
+            return waveforms, WaveformDataset._pack_metadata(metadata)
+        else:
+            return waveforms
+
+    @staticmethod
+    def _pack_metadata(metadata):
+        """
+        Reformats a list of dict into a dict of lists. Assumes identical keys in all dicts!
+        """
+
+        return {key: [m[key] for m in metadata] for key in metadata[0].keys()}
+
+    def _verify_grouping_defined(self):
+        """
+        Check if grouping is defined and raises and error otherwise
+        """
+        if self.grouping is None:
+            raise ValueError(
+                "Groups need to be defined first by assigning a value to grouping."
+            )
+
     def _get_single_waveform(
         self,
         trace_name,
@@ -1199,8 +1381,8 @@ class WaveformDataset:
         fig = plt.figure(figsize=(15, 10))
         try:
             import cartopy.crs as ccrs
-            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
             import cartopy.feature as cfeature
+            from cartopy.mpl.gridliner import LATITUDE_FORMATTER, LONGITUDE_FORMATTER
         except ModuleNotFoundError:
             raise ModuleNotFoundError(
                 "Plotting the data set requires cartopy. "
@@ -1329,6 +1511,8 @@ class MultiWaveformDataset:
         )
 
         self._homogenize_dataformat(datasets)
+        self._grouping = None
+        self._homogenize_grouping(datasets)
         self._build_trace_name_to_idx_dict()
 
     def __add__(self, other):
@@ -1399,6 +1583,54 @@ class MultiWaveformDataset:
             )
             self.missing_components = self.datasets[0].missing_components
 
+    def _homogenize_grouping(self, datasets):
+        groupings = [dataset.grouping for dataset in datasets]
+        if any(grouping != groupings[0] for grouping in groupings):
+            seisbench.logger.warning(
+                "Found inconsistent groupings. Setting grouping to None."
+            )
+            self.grouping = None
+        else:
+            self.grouping = groupings[0]
+
+    @property
+    def grouping(self):
+        """
+        The grouping parameters for the dataset.
+        Grouping allows to access metadata and waveforms
+        jointly from a set of traces with a common metadata parameter.
+        This can for example be used to access all waveforms belong to one event
+        and building event based models.
+        Setting the grouping parameter defines the output of
+        :py:attr:`~groups` and the associated methods.
+        `grouping` can be either a single string or a list of strings.
+        Each string must be a column in the metadata.
+        By default, the grouping is None.
+        """
+        return self._grouping
+
+    @grouping.setter
+    def grouping(self, value):
+        self._grouping = value
+        if value is None:
+            self._groups = None
+            self._groups_to_trace_idx = None
+        else:
+            self._groups_to_trace_idx = self.metadata.groupby(value).groups
+            self._groups = list(self._groups_to_trace_idx.keys())
+            self._groups_to_group_idx = {
+                group: i for i, group in enumerate(self._groups)
+            }
+
+    @property
+    def groups(self):
+        """
+        The list of groups as defined by the :py:attr:`~grouping` or `None` if :py:attr:`~grouping` is `None`.
+        """
+        return copy.copy(
+            self._groups
+        )  # Return a copy to make the internal groups immutable
+
     @property
     def datasets(self):
         """
@@ -1611,6 +1843,7 @@ class MultiWaveformDataset:
             # Calculate new metadata
             self._metadata = pd.concat(x.metadata for x in self.datasets)
             self._build_trace_name_to_idx_dict()
+            self.grouping = self.grouping  # Rebuild groups
 
         else:
             return MultiWaveformDataset(
@@ -1678,6 +1911,12 @@ class MultiWaveformDataset:
     train_dev_test = WaveformDataset.train_dev_test
     _build_trace_name_to_idx_dict = WaveformDataset._build_trace_name_to_idx_dict
     get_idx_from_trace_name = WaveformDataset.get_idx_from_trace_name
+    get_group_idx_from_params = WaveformDataset.get_group_idx_from_params
+    _verify_grouping_defined = WaveformDataset._verify_grouping_defined
+    get_group_waveforms = WaveformDataset.get_group_waveforms
+    get_group_samples = WaveformDataset.get_group_samples
+    get_group_size = WaveformDataset.get_group_size
+    _get_group_internal = WaveformDataset._get_group_internal
 
 
 class LoadingContext:
@@ -1752,6 +1991,8 @@ class BenchmarkDataset(WaveformDataset, ABC):
 
         if chunks is None:
             chunks = self.available_chunks(force=force, wait_for_file=wait_for_file)
+        else:
+            self._validate_chunks(self.path, chunks)
 
         for chunk in chunks:
 
@@ -1822,7 +2063,7 @@ class BenchmarkDataset(WaveformDataset, ABC):
         """
         Path to the dataset location in the SeisBench cache. This class method is required for technical reasons.
         """
-        return Path(seisbench.cache_root, "datasets", cls._name_internal().lower())
+        return Path(seisbench.cache_data_root, cls._name_internal().lower())
 
     @property
     def path(self):
@@ -1851,9 +2092,7 @@ class BenchmarkDataset(WaveformDataset, ABC):
         Path within the remote repository. Does only generate the pass without checking actual availability.
         Can be overwritten for datasets stored in the correct format but at a different location.
         """
-        return os.path.join(
-            seisbench.remote_root, "datasets", cls._name_internal().lower()
-        )
+        return urljoin(seisbench.remote_data_root, cls._name_internal().lower())
 
     @classmethod
     def available_chunks(cls, force=False, wait_for_file=False):
